/*
 * Copyright (c) 2005 Mellanox Technologies.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: ip_addresses_shared.h 47 2005-05-30 18:00:40Z sleybo $
 */

// This file is shared between the user and kernel

#ifndef _IP_ADDRESSES_SHARED_H_
#define _IP_ADDRESSES_SHARED_H_

#define IPOIB_IOCTL_VERSION  1

#define HW_ADDR_LEN		6


// This IRP is used to return all available CAs ports number and port guid
#define IOCTL_IPOIB_PORTS CTL_CODE(FILE_DEVICE_UNKNOWN, 0x801, METHOD_BUFFERED ,FILE_ANY_ACCESS)

struct IPOIB_AT_PORT_RECORD {
	ib_net64_t      CaGuid;
	ib_net64_t		PortGuid;
    ULONG           PortNumber;
};

struct IOCTL_IPOIB_PORTS_IN {
    ULONG				Version;
	ULONG				Size; //Total size, of the output buffers
};

struct IOCTL_IPOIB_PORTS_OUT {
    ULONG				    Version;
    ULONG				    Size; //Total size, of the output buffer needed if the
	                              // suplied buffer wasn't enough
    ULONG				    NumPorts;
    struct IPOIB_AT_PORT_RECORD    Ports[1];
};

// This IRP is used to return all the ip addresses that are assigned to a port
#define IOCTL_IPOIB_IP_ADDRESSES CTL_CODE(FILE_DEVICE_UNKNOWN, 0x802, METHOD_BUFFERED ,FILE_ANY_ACCESS)

struct IOCTL_IPOIB_IP_ADDRESSES_IN {
    ULONG				    Version;
	ib_net64_t		        PortGuid; // The guid of the port that we are queriyn for
};

struct IP_ADDRESS {
    char                    IpVersion;  // might only be 4 or 6
    unsigned char           Data[16];   // This allows to return ip v6 as well
};

struct IOCTL_IPOIB_IP_ADDRESSES_OUT {
    ULONG				    Version;
	ULONG				    Size; //Total size, of the output buffer needed if the
	                              // suplied buffer wasn't enough
	ULONG				    NumIps;
    struct IP_ADDRESS       Addreses[1];
};


// This IRP is used to convert a remote mac addresses to a remote GID
#define IOCTL_IPOIB_MAC_2_GID CTL_CODE(FILE_DEVICE_UNKNOWN, 0x803, METHOD_BUFFERED ,FILE_ANY_ACCESS)


struct IOCTL_IPOIB_MAC_2_GID_IN {
    unsigned char           DestMac[HW_ADDR_LEN];
};

struct IOCTL_IPOIB_MAC_2_GID_OUT {
    ib_gid_t                DestGid;
};

// The following structure functions and structurs define the functions that are needed in order to
// hold an array of the ports that are currently available.

#define DevName L"IPOIB"

#define IPOIB_DEV_NAME	L"\\Device\\"DevName
#define IPOIB_DOS_DEV_NAME L"\\DosDevices\\Global\\"DevName
#define IPOIB_WIN32_NAME L"\\\\.\\"DevName



#endif	/* _IP_ADDRESSES_SHARED_H_ */
