/*
 * Copyright (c) 2005 SilverStorm Technologies.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: ipoib_debug.h 1225 2006-04-05 08:01:01Z tzachid $
 */


#ifndef _IPOIB_DEBUG_H_
#define _IPOIB_DEBUG_H_


#include <complib/cl_debug.h>


/* Object types for passing into complib. */
#define IPOIB_OBJ_INSTANCE		1
#define IPOIB_OBJ_PORT			2
#define IPOIB_OBJ_ENDPOINT		3


/*
 * Debug macros
 */
extern uint32_t		g_ipoib_dbg_lvl;


#define IPOIB_DBG_INIT	(1 << 0)
#define IPOIB_DBG_PNP	(1 << 1)
#define IPOIB_DBG_SEND	(1 << 2)
#define IPOIB_DBG_RECV	(1 << 3)
#define IPOIB_DBG_ENDPT	(1 << 4)
#define IPOIB_DBG_IB	(1 << 5)
#define IPOIB_DBG_BUF	(1 << 6)
#define IPOIB_DBG_MCAST	(1 << 7)
#define IPOIB_DBG_ALLOC	(1 << 8)
#define IPOIB_DBG_OID	(1 << 9)
#define IPOIB_DBG_IOCTL	(1 << 10)
#define IPOIB_DBG_VM	(1 << 11)


#define IPOIB_DBG_FUNC	(1 << 28)	/* For function entry/exit */
#define IPOIB_DBG_INFO	(1 << 29)	/* For verbose information */
#define IPOIB_DBG_WARN	(1 << 30)	/* For warnings. */
#define IPOIB_DBG_ERROR	CL_DBG_ERROR
#define IPOIB_DBG_ALL	CL_DBG_ALL

/* Enter and exit macros automatically add IPOIB_DBG_FUNC bit */
#define IPOIB_ENTER( lvl )				\
	CL_ENTER( (lvl | IPOIB_DBG_FUNC), g_ipoib_dbg_lvl )
#define IPOIB_EXIT( lvl )				\
	CL_EXIT( (lvl | IPOIB_DBG_FUNC), g_ipoib_dbg_lvl )
#define IPOIB_TRACE( lvl, msg )			\
	CL_TRACE( (lvl), g_ipoib_dbg_lvl, msg )
#define IPOIB_TRACE_EXIT( lvl, msg )	\
	CL_TRACE_EXIT( (lvl), g_ipoib_dbg_lvl, msg )

#define IPOIB_TRACE_BYTES( lvl, ptr, len )									\
	{																		\
		size_t _loop_;														\
		for (_loop_ = 0; _loop_ < (len); ++_loop_)							\
		{																	\
			CL_PRINT( (lvl), g_ipoib_dbg_lvl, ("0x%.2X ", ((uint8_t*)(ptr))[_loop_]));	\
			if ((_loop_  + 1)% 16 == 0)											\
			{																\
				CL_PRINT( (lvl), g_ipoib_dbg_lvl, ("\n") );					\
			}																\
			else if ((_loop_ % 4 + 1) == 0)										\
			{																\
				CL_PRINT( (lvl), g_ipoib_dbg_lvl, ("  ") );					\
			}																\
		}																	\
		CL_PRINT( (lvl), g_ipoib_dbg_lvl, ("\n") );							\
	}


enum ipoib_perf_counters
{
	SendBundle,
	SendPackets,
		PortSend,
			GetEthHdr,
			SendMgrQueue,
				GetEndpt,
					EndptQueue,
			QueuePacket,
			BuildSendDesc,
				SendMgrFilter,
					FilterIp,
						QueryIp,
						SendTcp,
						FilterUdp,
							QueryUdp,
							SendUdp,
							FilterDhcp,
					FilterArp,
					SendGen,
						SendCopy,
			PostSend,
			ProcessFailedSends,
	SendCompBundle,
	SendCb,
		PollSend,
		SendComp,
		FreeSendBuf,
		RearmSend,
		PortResume,
	RecvCompBundle,
	RecvCb,
		PollRecv,
		FilterRecv,
			GetRecvEndpts,
				GetEndptByGid,
				GetEndptByLid,
				EndptInsert,
			RecvTcp,
			RecvUdp,
			RecvDhcp,
			RecvArp,
			RecvGen,
	BuildPktArray,
		PreparePkt,
			GetNdisPkt,
	RecvNdisIndicate,
	PutRecvList,
	RepostRecv,
		GetRecv,
		PostRecv,
	RearmRecv,
		ReturnPacket,
		ReturnPutRecv,
		ReturnRepostRecv,
		ReturnPreparePkt,
		ReturnNdisIndicate,

	/* Must be last! */
	MaxPerf

};

#endif	/* _IPOIB_DEBUG_H_ */
